package com.Ekyc.ekyc2022.camerax

import android.annotation.SuppressLint
import android.media.MediaScannerConnection
import android.util.Size
import android.view.MotionEvent
import android.view.View
import android.widget.Toast
import androidx.appcompat.app.AppCompatActivity
import androidx.camera.core.*
import androidx.camera.lifecycle.ProcessCameraProvider
import androidx.camera.view.PreviewView
import com.Ekyc.ekyc2022.global.EKYCManager
import java.io.File
import java.util.*
import java.util.concurrent.Executor
import kotlin.math.roundToInt

@Suppress("UNUSED")
class CameraManager private constructor(
    private val activity: AppCompatActivity,
    private val cameraPreviewView: PreviewView,
    private val cameraMode: CameraMode,
    private val cameraExecutor: Executor,
    private val callback: CameraCallback?,
    private var _cameraSelectorOption: Int,
    private val aspectRatio: Float
) {
    companion object {
        private const val PHOTO_FILENAME_FORMAT = "image_%d.jpg"
        private const val VIDEO_FILENAME_FORMAT = "video_%d.mp4"
        private const val DEFAULT_ASPECT_RATIO = 4f / 3
        private const val HD_RESOLUTION_WIDTH_PX = 1080
        private const val NOTIFY_FILE_INSERTED = false
    }

    private var cameraProvider: ProcessCameraProvider? = null
    val cameraSelectorOption: Int
        get() = _cameraSelectorOption
    private var _flashMode = ImageCapture.FLASH_MODE_OFF
    val flashMode: Int
        get() = _flashMode
    private var imageCapture: ImageCapture? = null
    private var videoCapture: VideoCapture? = null
    private var isRecording = false
    private var outputFileDirectory: File

    init {
        outputFileDirectory = getOutputDirectory()
    }

    @Throws(Exception::class)
    fun startCamera() {
        EKYCManager.checkInitialized()
        reset()
        val cameraProviderFuture = ProcessCameraProvider.getInstance(activity)
        cameraProviderFuture.addListener(
            {
                cameraProvider = cameraProviderFuture.get()
                val cameraSelector = CameraSelector.Builder()
                    .requireLensFacing(_cameraSelectorOption)
                    .build()

                cameraProvider?.unbindAll()

                try {
                    val camera = cameraProvider?.bindToLifecycle(
                        activity,
                        cameraSelector,
                        buildCameraUseCases()
                    )

                    setupTapToFocus(camera!!.cameraControl)
                } catch (e: Exception) {
                    activity.runOnUiThread {
                        Toast.makeText(
                            activity,
                            "Không tìm thấy camera, vui lòng kiểm tra lại camera hoặc khởi động lại ứng dụng.",
                            Toast.LENGTH_SHORT
                        ).show()
                    }
                }
            }, cameraExecutor
        )
    }

    private fun reset() {
        imageCapture = null
        videoCapture = null
    }

    fun doAction() {
        if (cameraMode == CameraMode.MODE_PHOTO) {
            takePhoto()
        } else {
            toggleRecord()
        }
    }

    private fun takePhoto() {
        if (imageCapture != null) {
            imageCapture!!.flashMode = _flashMode

            val photoFile = File(
                outputFileDirectory,
                String.format(PHOTO_FILENAME_FORMAT, System.currentTimeMillis())
            )
            val metadata = ImageCapture.Metadata()
            metadata.isReversedHorizontal =
                _cameraSelectorOption == CameraSelector.LENS_FACING_FRONT
            val outputOptions =
                ImageCapture.OutputFileOptions.Builder(photoFile).setMetadata(metadata).build()

            imageCapture!!.takePicture(
                outputOptions,
                cameraExecutor,
                object : ImageCapture.OnImageSavedCallback {
                    override fun onImageSaved(output: ImageCapture.OutputFileResults) {
                        callback?.onSuccess(photoFile.absolutePath)
                        notifyMediaInserted(photoFile)
                    }

                    override fun onError(exception: ImageCaptureException) {
                        callback?.onError(exception.message)
                    }
                })
        }
    }

    private fun toggleRecord() {
        if (isRecording) {
            stopRecording()
        } else {
            startRecording()
        }
    }

    @SuppressLint("MissingPermission", "RestrictedApi")
    private fun startRecording() {
        if (videoCapture != null && !isRecording) {
            val outputFile = File(
                outputFileDirectory,
                String.format(VIDEO_FILENAME_FORMAT, System.currentTimeMillis())
            )

            videoCapture!!.startRecording(
                VideoCapture.OutputFileOptions.Builder(outputFile).build(),
                cameraExecutor,
                object : VideoCapture.OnVideoSavedCallback {
                    override fun onVideoSaved(outputFileResults: VideoCapture.OutputFileResults) {
                        callback?.onSuccess(outputFile.absolutePath)
                        notifyMediaInserted(outputFile)
                    }

                    override fun onError(
                        videoCaptureError: Int,
                        message: String,
                        cause: Throwable?
                    ) {
                        callback?.onError(message)
                        try {
                            stopRecording()
                        } catch (e: Exception) {
                        }
                    }
                }
            )
            isRecording = true
        }
    }

    private fun notifyMediaInserted(file: File) {
        if (NOTIFY_FILE_INSERTED) {
            MediaScannerConnection.scanFile(activity, arrayOf(file.toString()), null, null)
        }
    }

    @SuppressLint("RestrictedApi")
    private fun stopRecording() {
        if (videoCapture != null) {
            videoCapture!!.stopRecording()
        }
        isRecording = false
    }

    private fun setupTapToFocus(cameraControl: CameraControl) {
        cameraPreviewView.setOnTouchListener(
            View.OnTouchListener setOnTouchListener@{ _: View, motionEvent: MotionEvent ->
                when (motionEvent.action) {
                    MotionEvent.ACTION_DOWN -> return@setOnTouchListener true
                    MotionEvent.ACTION_UP -> {
                        val factory = cameraPreviewView.meteringPointFactory
                        val point = factory.createPoint(motionEvent.x, motionEvent.y)
                        val action = FocusMeteringAction.Builder(point).build()
                        cameraControl.startFocusAndMetering(action)
                        return@setOnTouchListener true
                    }
                    else -> return@setOnTouchListener false
                }
            })
    }

    @SuppressLint("RestrictedApi")
    private fun buildCameraUseCases(): UseCaseGroup {
        val preview = Preview.Builder().build()
            .also { preview -> preview.setSurfaceProvider(cameraPreviewView.surfaceProvider) }
        val useCaseGroupBuilder = UseCaseGroup.Builder().addUseCase(preview)
        val hdResolutionWidth = HD_RESOLUTION_WIDTH_PX
        val maxResolution = Size(
            hdResolutionWidth,
            ((hdResolutionWidth / aspectRatio).roundToInt())
        )

        return when (cameraMode) {
            CameraMode.MODE_PHOTO -> {
                imageCapture = ImageCapture.Builder()
                    .setTargetResolution(maxResolution)
                    .setFlashMode(_flashMode)
                    .build()
                useCaseGroupBuilder
                    .addUseCase(imageCapture!!)
                    .build()
            }
            CameraMode.MODE_VIDEO -> {
                videoCapture = VideoCapture.Builder()
                    .setTargetResolution(maxResolution)
                    .build()
                useCaseGroupBuilder
                    .addUseCase(videoCapture!!)
                    .build()
            }
        }
    }

    fun toggleCameraSelector(): Int {
        cameraProvider?.unbindAll()
        _cameraSelectorOption =
            if (_cameraSelectorOption == CameraSelector.LENS_FACING_BACK) CameraSelector.LENS_FACING_FRONT
            else CameraSelector.LENS_FACING_BACK
        startCamera()
        return _cameraSelectorOption
    }

    private fun getOutputDirectory(): File {
        val mediaDir = activity.externalMediaDirs.firstOrNull()?.let {
            File(it, "EKYC").apply { mkdirs() }
        }
        return if (mediaDir != null && mediaDir.exists())
            mediaDir else activity.filesDir
    }

    fun changeFlashMode(@ImageCapture.FlashMode mode: Int): Int {
        _flashMode = mode
        return mode
    }

    interface CameraCallback {
        fun onSuccess(filePath: String)
        fun onError(errorMessage: String?)
    }

    class Builder {
        private var activity: AppCompatActivity? = null
        private var cameraPreviewView: PreviewView? = null
        private var cameraMode: CameraMode? = null
        private var cameraExecutor: Executor? = null
        private var callback: CameraCallback? = null
        private var cameraSelectorOption = CameraSelector.LENS_FACING_BACK
        private var aspectRatio = DEFAULT_ASPECT_RATIO

        fun setActivity(activity: AppCompatActivity): Builder {
            this.activity = activity
            return this
        }

        /**
         * @param cameraPreviewView PreviewView for camera.
         * The camera's aspect ratio is [aspectRatio] and it doesn't depend on PreviewView's size.
         */
        fun setCameraPreview(cameraPreviewView: PreviewView): Builder {
            this.cameraPreviewView = cameraPreviewView
            return this
        }

        fun setCameraMode(cameraMode: CameraMode): Builder {
            this.cameraMode = cameraMode
            cameraSelectorOption =
                if (cameraMode == CameraMode.MODE_PHOTO)
                    CameraSelector.LENS_FACING_BACK else CameraSelector.LENS_FACING_FRONT
            return this
        }

        fun setCameraExecutor(cameraExecutor: Executor): Builder {
            this.cameraExecutor = cameraExecutor
            return this
        }

        fun setCallback(callback: CameraCallback): Builder {
            this.callback = callback
            return this
        }

        fun setCameraSelectorOption(cameraSelectorOption: CameraOption): Builder {
            this.cameraSelectorOption =
                if (cameraSelectorOption == CameraOption.LENS_FACING_FRONT)
                    CameraSelector.LENS_FACING_FRONT else
                    CameraSelector.LENS_FACING_BACK
            return this
        }

        fun setAspectRatio(aspectRatio: Float): Builder {
            this.aspectRatio = aspectRatio
            return this
        }

        fun build(): CameraManager {
            checkNotNull(activity)
            checkNotNull(cameraPreviewView)
            checkNotNull(cameraMode)
            checkNotNull(cameraExecutor)

            return CameraManager(
                activity!!,
                cameraPreviewView!!,
                cameraMode!!,
                cameraExecutor!!,
                callback,
                cameraSelectorOption,
                aspectRatio
            )
        }
    }
}