package com.Ekyc.ekyc2022.facedetection

import android.graphics.Bitmap
import android.graphics.Rect
import android.util.Log
import androidx.camera.view.PreviewView
import com.google.android.gms.tasks.Task
import com.google.mlkit.vision.common.InputImage
import com.google.mlkit.vision.face.Face
import com.google.mlkit.vision.face.FaceDetection
import com.google.mlkit.vision.face.FaceDetectorOptions
import com.Ekyc.ekyc2022.global.Utils.releaseBitmap
import java.io.IOException
import kotlin.math.abs

@Suppress("UNUSED")
class FaceContourDetectionProcessor(
    private val callback: Callback,
    private val borderMarginHorizontal: Float = 0f,
    private val borderMarginVertical: Float = 0f,
) {
    companion object {
        private const val TAG = "FaceDetectorProcessor"
    }

    private val realTimeOpts = FaceDetectorOptions.Builder()
        .setPerformanceMode(FaceDetectorOptions.PERFORMANCE_MODE_FAST)
        .setContourMode(FaceDetectorOptions.CONTOUR_MODE_NONE)
        .setClassificationMode(FaceDetectorOptions.CLASSIFICATION_MODE_ALL)
        .build()
    private val detector = FaceDetection.getClient(realTimeOpts)
    private var isProcessing = false

    private fun detectInImage(image: InputImage): Task<List<Face>> {
        return detector.process(image)
    }

    fun stop() {
        try {
            isProcessing = false
            detector.close()
        } catch (e: IOException) {
            Log.e(TAG, "Exception thrown while trying to close Face Detector: $e")
        }
    }

    fun analyze(previewView: PreviewView) {
        analyze(previewView.bitmap)
    }

    @SuppressWarnings("WeakerAccess")
    @Synchronized
    fun analyze(bitmap: Bitmap?) {
        bitmap?.let {
            if (!isProcessing) {
                isProcessing = true
                val mediaImage = InputImage.fromBitmap(it, 0 /* rotationDegrees */)
                detectInImage(mediaImage)
                    .addOnSuccessListener { results ->
                        onSuccess(bitmap, results, Rect(0, 0, bitmap.width, bitmap.height))
                        reset()
                    }
                    .addOnFailureListener { exception ->
                        onFailure(exception)
                        reset(it)
                    }
            } else {
                bitmap.releaseBitmap()
            }
            return
        }
    }

    private fun reset(bitmap: Bitmap? = null) {
        isProcessing = false
        if (bitmap?.isRecycled == false) {
            bitmap.recycle()
        }
    }

    private fun onSuccess(
        bitmap: Bitmap,
        results: List<Face>,
        border: Rect
    ) {
        if (results.isNullOrEmpty()) {
            callback.onAnalyzed(bitmap, AnalyzeResult.FAILED)
        } else if (results.size == 1) {
            val face = results[0]
            if (!checkFacePosition(face.boundingBox, border)) {
                callback.onAnalyzed(bitmap, AnalyzeResult.INVALID_POSITION)
            } else if (!checkFacialPosture(face)) {
                callback.onAnalyzed(bitmap, AnalyzeResult.INVALID_FACIAL_POSTURE)
            } else {
                callback.onAnalyzed(bitmap, AnalyzeResult.OK)
            }
        } else {
            callback.onAnalyzed(bitmap, AnalyzeResult.MULTIPLE_FACES)
        }
    }

    private fun area(rect: Rect): Int {
        return rect.width() * rect.height()
    }

    private fun checkFacePosition(faceBoundingBox: Rect, border: Rect): Boolean {
        return faceBoundingBox.left > borderMarginHorizontal
                && faceBoundingBox.top > borderMarginVertical
                && faceBoundingBox.right < border.right - borderMarginHorizontal
                && faceBoundingBox.bottom < border.bottom - borderMarginVertical
    }

    private fun checkFacialPosture(face: Face): Boolean {
        return abs(face.headEulerAngleX) < 20
                && abs(face.headEulerAngleY) < 20
                && abs(face.headEulerAngleZ) < 20
    }

    private fun checkEyesOpenProbability(face: Face): Boolean {
        return (face.leftEyeOpenProbability == null || face.leftEyeOpenProbability!! >= 0.2f)
                && (face.rightEyeOpenProbability == null || face.rightEyeOpenProbability!! >= 0.2f)
    }

    private fun onFailure(e: Exception) {
        Log.w(TAG, "Face Detector failed.$e")
    }

    interface Callback {
        fun onAnalyzed(bitmap: Bitmap, analyzeResult: AnalyzeResult)
    }
}