package com.Ekyc.ekyc2022.objectdetection

import android.graphics.Bitmap
import org.opencv.android.Utils
import org.opencv.calib3d.Calib3d
import org.opencv.calib3d.Calib3d.RANSAC
import org.opencv.core.*
import org.opencv.imgproc.Imgproc
import kotlin.math.acos
import kotlin.math.pow
import kotlin.math.sin
import kotlin.math.sqrt

object AlignIdImageHelper {
    private const val LOW_BLUR_THRESHOLD = 30.0
    private const val BLUR_THRESHOLD = 120.0
    private const val ENABLE_BLUR_CHECKING = false
    private const val ENABLE_PADDINGS = false
    const val FRONT_ID_CARD_POINTS = 5 // 4 corners and emblem
    const val BACK_ID_CARD_POINTS = 7 // 4 corners, symbol and fingerprints.

    fun alignIDCard(bitmap: Bitmap, points: List<Point>, type: String): Bitmap? {
        return when (type) {
            IdCardType.BACK_IDENTITY_CARD -> alignBackCMND(bitmap, points)
            IdCardType.BACK_CITIZEN_IDENTIFICATION -> alignBackCCCD(bitmap, points)
            IdCardType.BACK_CHIP_BASED_CITIZEN_IDENTIFICATION -> alignBackCHIP(bitmap, points)
            else -> alignFrontIDCard(bitmap, points, type)
        }
    }

    private fun alignFrontIDCard(bitmap: Bitmap, points: List<Point>, type: String): Bitmap? {
        if (points.size != FRONT_ID_CARD_POINTS) {
            return bitmap
        }

        //Get image size
        val imgWidth = bitmap.width
        val imgHeight = imgWidth * 63 / 84

        val srcMat = Mat()
        val destMat = Mat(Size(imgWidth.toDouble(), imgHeight.toDouble()), RANSAC)
        Utils.bitmapToMat(bitmap, srcMat)

        val srcObj = MatOfPoint2f()
        val destObj = MatOfPoint2f()

        //Sort corner
        val realSymbolX: Double = points[4].x// * srcMat.width().toDouble()
        val realSymbolY: Double = points[4].y// * srcMat.height().toDouble()

        val realX1: Double = points[0].x// * srcMat.width().toDouble()
        val realX2: Double = points[1].x// * srcMat.width().toDouble()
        val realX3: Double = points[2].x// * srcMat.width().toDouble()
        val realX4: Double = points[3].x// * srcMat.width().toDouble()

        val realY1: Double = points[0].y// * srcMat.height().toDouble()
        val realY2: Double = points[1].y// * srcMat.height().toDouble()
        val realY3: Double = points[2].y// * srcMat.height().toDouble()
        val realY4: Double = points[3].y// * srcMat.height().toDouble()

        val vec1: Double = sqrt((realX1 - realSymbolX).pow(2) + (realY1 - realSymbolY).pow(2))
        val vec2: Double = sqrt((realX2 - realSymbolX).pow(2) + (realY2 - realSymbolY).pow(2))
        val vec3: Double = sqrt((realX3 - realSymbolX).pow(2) + (realY3 - realSymbolY).pow(2))
        val vec4: Double = sqrt((realX4 - realSymbolX).pow(2) + (realY4 - realSymbolY).pow(2))
        val idPoints: List<IDPoint> = listOf(
            IDPoint(vec1, Point(realX1, realY1)),
            IDPoint(vec2, Point(realX2, realY2)),
            IDPoint(vec3, Point(realX3, realY3)),
            IDPoint(vec4, Point(realX4, realY4))
        ).sortedBy { it.vec }

        val point1 = idPoints[0]
        val point2 = idPoints[2]
        val point3 = idPoints[3]
        val point4 = idPoints[1]

        val listPoint = arrayListOf(point1, point2, point3, point4)
        paddingImage(listPoint)

        //Find the transformation matrix
        srcObj.fromArray(
            listPoint[0].point,
            listPoint[1].point,
            listPoint[2].point,
            listPoint[3].point
        )
        destObj.fromArray(
            Point(0.0, 0.0),
            Point(imgWidth.toDouble(), 0.0),
            Point(imgWidth.toDouble(), imgHeight.toDouble()),
            Point(0.0, imgHeight.toDouble())
        )

        val m = Calib3d.findHomography(srcObj, destObj, RANSAC)
        Imgproc.warpPerspective(srcMat, destMat, m, destMat.size())

        //Create new bitmap
        val newBitmap = Bitmap.createBitmap(destMat.cols(), destMat.rows(), Bitmap.Config.ARGB_8888)
        Utils.matToBitmap(destMat, newBitmap)
        return newBitmap
    }

    private fun alignBackCMND(bitmap: Bitmap, points: List<Point>): Bitmap {
        //Get image size
        val imgWidth = bitmap.width
        val imgHeight = imgWidth * 63 / 84

        val srcMat = Mat()
        val destMat = Mat(Size(imgWidth.toDouble(), imgHeight.toDouble()), RANSAC)
        Utils.bitmapToMat(bitmap, srcMat)

        val srcObj = MatOfPoint2f()
        val destObj = MatOfPoint2f()

        //Sort corner
        val checkPoint = Point(
            (points[4].x + points[5].x + points[6].x) / 3,
            (points[4].y + points[5].y + points[6].y) / 3
        )

        val realSymbolX: Double = checkPoint.x// * srcMat.width().toDouble()
        val realSymbolY: Double = checkPoint.y// * srcMat.width().toDouble()

        val realX1: Double = points[0].x// * srcMat.width().toDouble()
        val realX2: Double = points[1].x// * srcMat.width().toDouble()
        val realX3: Double = points[2].x// * srcMat.width().toDouble()
        val realX4: Double = points[3].x// * srcMat.width().toDouble()
        val realX5: Double = points[4].x// * srcMat.width().toDouble()
        val realX7: Double = points[6].x// * srcMat.width().toDouble()

        val realY1: Double = points[0].y// * srcMat.height().toDouble()
        val realY2: Double = points[1].y// * srcMat.height().toDouble()
        val realY3: Double = points[2].y// * srcMat.height().toDouble()
        val realY4: Double = points[3].y// * srcMat.height().toDouble()
        val realY5: Double = points[4].y// * srcMat.height().toDouble()
        val realY7: Double = points[6].y// * srcMat.height().toDouble()

        val vec1: Double = sqrt((realX1 - realSymbolX).pow(2) + (realY1 - realSymbolY).pow(2))
        val vec2: Double = sqrt((realX2 - realSymbolX).pow(2) + (realY2 - realSymbolY).pow(2))
        val vec3: Double = sqrt((realX3 - realSymbolX).pow(2) + (realY3 - realSymbolY).pow(2))
        val vec4: Double = sqrt((realX4 - realSymbolX).pow(2) + (realY4 - realSymbolY).pow(2))
        val idPoints: List<IDPoint> = listOf(
            IDPoint(vec1, Point(realX1, realY1)),
            IDPoint(vec2, Point(realX2, realY2)),
            IDPoint(vec3, Point(realX3, realY3)),
            IDPoint(vec4, Point(realX4, realY4))
        ).sortedBy { it.vec }

        var point1 = idPoints[1]
        var point2 = idPoints[3]
        var point3 = idPoints[2]
        var point4 = idPoints[0]

        var x1: Double = point1.point.x
        var y1: Double = point1.point.y
        var x2: Double = point2.point.x
        var y2: Double = point2.point.y
        var x3: Double = point3.point.x
        var y3: Double = point3.point.y
        var x4: Double = point4.point.x
        var y4: Double = point4.point.y

        //Check flip vertical
        val ch1 = sqrt((realX7 - x1).pow(2) + (realY7 - y1).pow(2))
        val vector1 = Point(x2 - x1, y2 - y1)
        val vector2 = Point(realX7 - x1, realY7 - y1)
        val ts1 = vector1.x * vector2.x + vector1.y * vector2.y
        val ms1 =
            sqrt(vector1.x * vector1.x + vector1.y * vector1.y) * sqrt(vector2.x * vector2.x + vector2.y * vector2.y)
        val rad1 = acos(ts1 / ms1)
        val sinCheck1 = sin(rad1)
        val d1 = sinCheck1 * ch1

        val ch2 = sqrt((realX7 - x3).pow(2) + (realY7 - y3).pow(2))
        val vector3 = Point(x3 - x4, y3 - y4)
        val vector4 = Point(realX7 - x4, realY7 - y4)
        val ts2 = vector3.x * vector4.x + vector3.y * vector4.y
        val ms2 =
            sqrt(vector3.x * vector3.x + vector3.y * vector3.y) * sqrt(vector4.x * vector4.x + vector4.y * vector4.y)
        val rad2 = acos(ts2 / ms2)
        val sinCheck2 = sin(rad2)
        val d2 = sinCheck2 * ch2

        if (d1 < d2) {
            point1 = idPoints[0]
            point2 = idPoints[2]
            point3 = idPoints[3]
            point4 = idPoints[1]

            x1 = point1.point.x
            y1 = point1.point.y
            x2 = point2.point.x
            y2 = point2.point.y
            x3 = point3.point.x
            y3 = point3.point.y
            x4 = point4.point.x
            y4 = point4.point.y
        }

        //Check flip horizontal
        val hch1 = sqrt((realX5 - x1).pow(2) + (realY5 - y1).pow(2))
        val hvector1 = Point(x4 - x1, y4 - y1)
        val hvector2 = Point(realX5 - x1, realY5 - y1)
        val hts1 = hvector1.x * hvector2.x + hvector1.y * hvector2.y
        val hms1 =
            sqrt(hvector1.x * hvector1.x + hvector1.y * hvector1.y) * sqrt(hvector2.x * hvector2.x + hvector2.y * hvector2.y)
        val hrad1 = acos(hts1 / hms1)
        val hsinCheck1 = sin(hrad1)
        val hd1 = hsinCheck1 * hch1

        val hch2 = sqrt((realX5 - x2).pow(2) + (realY5 - y2).pow(2))
        val hvector3 = Point(x3 - x2, y3 - y2)
        val hvector4 = Point(realX5 - x2, realY5 - y2)
        val hts2 = hvector3.x * hvector4.x + hvector3.y * hvector4.y
        val hms2 =
            sqrt(hvector3.x * hvector3.x + hvector3.y * hvector3.y) * sqrt(hvector4.x * hvector4.x + hvector4.y * hvector4.y)
        val hrad2 = acos(hts2 / hms2)
        val hsinCheck2 = sin(hrad2)
        val hd2 = hsinCheck2 * hch2

        if (hd1 > hd2) {
            x1 = point2.point.x
            y1 = point2.point.y
            x2 = point1.point.x
            y2 = point1.point.y
            x3 = point4.point.x
            y3 = point4.point.y
            x4 = point3.point.x
            y4 = point3.point.y
        }

        //Add padding
        val listPoint = arrayListOf(point1, point2, point3, point4)
        paddingImage(listPoint)

        //Find the transformation matrix
        srcObj.fromArray(
            listPoint[0].point,
            listPoint[1].point,
            listPoint[2].point,
            listPoint[3].point
        )

        destObj.fromArray(
            Point(0.0, 0.0),
            Point(imgWidth.toDouble(), 0.0),
            Point(imgWidth.toDouble(), imgHeight.toDouble()),
            Point(0.0, imgHeight.toDouble())
        )

        val m = Calib3d.findHomography(srcObj, destObj, RANSAC)
        Imgproc.warpPerspective(srcMat, destMat, m, destMat.size())

        //Create new bitmap
        val newBitmap = Bitmap.createBitmap(destMat.cols(), destMat.rows(), Bitmap.Config.ARGB_8888)
        Utils.matToBitmap(destMat, newBitmap)
        return newBitmap
    }

    private fun alignBackCCCD(bitmap: Bitmap, points: List<Point>): Bitmap {
        //Get image size
        val imgWidth = bitmap.width
        val imgHeight = imgWidth * 63 / 84

        val srcMat = Mat()
        val destMat = Mat(Size(imgWidth.toDouble(), imgHeight.toDouble()), RANSAC)
        Utils.bitmapToMat(bitmap, srcMat)

        val srcObj = MatOfPoint2f()
        val destObj = MatOfPoint2f()

        //Sort corner
        val realSymbolX: Double = points[6].x// * srcMat.width().toDouble()
        val realSymbolY: Double = points[6].y// * srcMat.height().toDouble()

        val realX1: Double = points[0].x// * srcMat.width().toDouble()
        val realX2: Double = points[1].x// * srcMat.width().toDouble()
        val realX3: Double = points[2].x// * srcMat.width().toDouble()
        val realX4: Double = points[3].x// * srcMat.width().toDouble()

        val realY1: Double = points[0].y// * srcMat.height().toDouble()
        val realY2: Double = points[1].y// * srcMat.height().toDouble()
        val realY3: Double = points[2].y// * srcMat.height().toDouble()
        val realY4: Double = points[3].y// * srcMat.height().toDouble()

        val vec1: Double = sqrt((realX1 - realSymbolX).pow(2) + (realY1 - realSymbolY).pow(2))
        val vec2: Double = sqrt((realX2 - realSymbolX).pow(2) + (realY2 - realSymbolY).pow(2))
        val vec3: Double = sqrt((realX3 - realSymbolX).pow(2) + (realY3 - realSymbolY).pow(2))
        val vec4: Double = sqrt((realX4 - realSymbolX).pow(2) + (realY4 - realSymbolY).pow(2))
        val idPoints: List<IDPoint> = listOf(
            IDPoint(vec1, Point(realX1, realY1)),
            IDPoint(vec2, Point(realX2, realY2)),
            IDPoint(vec3, Point(realX3, realY3)),
            IDPoint(vec4, Point(realX4, realY4))
        ).sortedBy { it.vec }

        val point1 = idPoints[3]
        val point2 = idPoints[1]
        val point3 = idPoints[0]
        val point4 = idPoints[2]

        var x1: Double = point1.point.x
        var y1: Double = point1.point.y
        var x2: Double = point2.point.x
        var y2: Double = point2.point.y
        var x3: Double = point3.point.x
        var y3: Double = point3.point.y
        var x4: Double = point4.point.x
        var y4: Double = point4.point.y

        //Add padding
        val listPoint = arrayListOf(point1, point2, point3, point4)
        paddingImage(listPoint)

        //Find the transformation matrix
        srcObj.fromArray(
            listPoint[0].point,
            listPoint[1].point,
            listPoint[2].point,
            listPoint[3].point
        )

        destObj.fromArray(
            Point(0.0, 0.0),
            Point(imgWidth.toDouble(), 0.0),
            Point(imgWidth.toDouble(), imgHeight.toDouble()),
            Point(0.0, imgHeight.toDouble())
        )

        val m = Calib3d.findHomography(srcObj, destObj, RANSAC)
        Imgproc.warpPerspective(srcMat, destMat, m, destMat.size())

        //Create new bitmap
        val newBitmap = Bitmap.createBitmap(destMat.cols(), destMat.rows(), Bitmap.Config.ARGB_8888)
        Utils.matToBitmap(destMat, newBitmap)
        return newBitmap
    }

    private fun alignBackCHIP(bitmap: Bitmap, points: List<Point>): Bitmap {
        //Get image size
        val imgWidth = bitmap.width
        val imgHeight = imgWidth * 63 / 84

        val srcMat = Mat()
        val destMat = Mat(Size(imgWidth.toDouble(), imgHeight.toDouble()), RANSAC)
        Utils.bitmapToMat(bitmap, srcMat)

        val srcObj = MatOfPoint2f()
        val destObj = MatOfPoint2f()

        //Sort corner
        val checkPoint = Point((points[4].x + points[5].x) / 2, (points[4].y + points[5].y) / 2)

        val realSymbolX: Double = checkPoint.x// * srcMat.width().toDouble()
        val realSymbolY: Double = checkPoint.y// * srcMat.width().toDouble()

        val realX1: Double = points[0].x// * srcMat.width().toDouble()
        val realX2: Double = points[1].x// * srcMat.width().toDouble()
        val realX3: Double = points[2].x// * srcMat.width().toDouble()
        val realX4: Double = points[3].x// * srcMat.width().toDouble()

        val realY1: Double = points[0].y// * srcMat.height().toDouble()
        val realY2: Double = points[1].y// * srcMat.height().toDouble()
        val realY3: Double = points[2].y// * srcMat.height().toDouble()
        val realY4: Double = points[3].y// * srcMat.height().toDouble()

        val vec1: Double = sqrt((realX1 - realSymbolX).pow(2) + (realY1 - realSymbolY).pow(2))
        val vec2: Double = sqrt((realX2 - realSymbolX).pow(2) + (realY2 - realSymbolY).pow(2))
        val vec3: Double = sqrt((realX3 - realSymbolX).pow(2) + (realY3 - realSymbolY).pow(2))
        val vec4: Double = sqrt((realX4 - realSymbolX).pow(2) + (realY4 - realSymbolY).pow(2))
        val idPoints: List<IDPoint> = listOf(
            IDPoint(vec1, Point(realX1, realY1)),
            IDPoint(vec2, Point(realX2, realY2)),
            IDPoint(vec3, Point(realX3, realY3)),
            IDPoint(vec4, Point(realX4, realY4))
        ).sortedBy { it.vec }

        val point1 = idPoints[2]
        val point2 = idPoints[0]
        val point3 = idPoints[1]
        val point4 = idPoints[3]

        val listPoint = arrayListOf(point1, point2, point3, point4)
        paddingImage(listPoint)

        //Find the transformation matrix
        srcObj.fromArray(
            listPoint[0].point,
            listPoint[1].point,
            listPoint[2].point,
            listPoint[3].point
        )

        destObj.fromArray(
            Point(0.0, 0.0),
            Point(imgWidth.toDouble(), 0.0),
            Point(imgWidth.toDouble(), imgHeight.toDouble()),
            Point(0.0, imgHeight.toDouble())
        )

        val m = Calib3d.findHomography(srcObj, destObj, RANSAC)
        Imgproc.warpPerspective(srcMat, destMat, m, destMat.size())

        //Create new bitmap
        val newBitmap = Bitmap.createBitmap(destMat.cols(), destMat.rows(), Bitmap.Config.ARGB_8888)
        Utils.matToBitmap(destMat, newBitmap)
        return newBitmap
    }

    fun isBlurredImage(bitmap: Bitmap, lowThreshold: Boolean = false): Boolean {
        if (!ENABLE_BLUR_CHECKING) {
            return false
        }

        val destination = Mat()
        val matGray = Mat()

        val rootWidth = bitmap.width.toDouble()
        val rootHeight = bitmap.height.toDouble()
        val rate: Double = rootWidth / rootHeight
        val newWidth = if (rootWidth >= rootHeight) 700.0 else 500.0
        val newHeight = newWidth / rate
        val newBitmap =
            Bitmap.createScaledBitmap(bitmap, newWidth.toInt(), newHeight.toInt(), false)
        val sourceMatImage = Mat()
        Utils.bitmapToMat(newBitmap, sourceMatImage)

        Imgproc.cvtColor(sourceMatImage, matGray, Imgproc.COLOR_BGR2GRAY)
        Imgproc.Laplacian(matGray, destination, CvType.CV_8U)
        val median = MatOfDouble()
        val std = MatOfDouble()
        Core.meanStdDev(destination, median, std)
        val score = std.get(0, 0)[0].pow(2.0)
        return score < if (lowThreshold) LOW_BLUR_THRESHOLD else BLUR_THRESHOLD
    }

    private fun paddingImage(points: List<IDPoint>) {
        if (ENABLE_PADDINGS) {
            var x1 = points[0].point.x
            var y1 = points[0].point.y
            var x2 = points[1].point.x
            var y2 = points[1].point.y
            var x3 = points[2].point.x
            var y3 = points[2].point.y
            var x4 = points[3].point.x
            var y4 = points[3].point.y

            //Add padding
            //Xác định đường thẳng đi qua point 1 3
            val a1: Double = y3 - y1
            val b1: Double = x1 - x3
            val c1: Double = y1 * x3 - x1 * y3

            //Xác định đường thẳng đi qua point 2 4
            val a2: Double = y4 - y2
            val b2: Double = x2 - x4
            val c2: Double = y2 * x4 - x2 * y4

            //Xác định giao điểm của 2 đường thẳng trên
            val xCenter = (b2 * c1 - b1 * c2) / (b1 * a2 - b2 * a1)
            val yCenter = (-c1 - a1 * xCenter) / b1

            val distance1 = sqrt((xCenter - x1) * (xCenter - x1) + (yCenter - y1) * (yCenter - y1))
            val distance2 = sqrt((xCenter - x2) * (xCenter - x2) + (yCenter - y2) * (yCenter - y2))
            val distance3 = sqrt((xCenter - x3) * (xCenter - x3) + (yCenter - y3) * (yCenter - y3))
            val distance4 = sqrt((xCenter - x4) * (xCenter - x4) + (yCenter - y4) * (yCenter - y4))

            val padding = distance1 / 3.0

            val ratePadding1 = distance1 / padding
            val ratePadding2 = distance2 / padding
            val ratePadding3 = distance3 / padding
            val ratePadding4 = distance4 / padding

            val dx1: Double = Math.abs(xCenter - x1) / ratePadding1
            val dy1: Double = Math.abs(yCenter - y1) / ratePadding1
            val dx2: Double = Math.abs(xCenter - x2) / ratePadding2
            val dy2: Double = Math.abs(yCenter - y2) / ratePadding2
            val dx3: Double = Math.abs(xCenter - x3) / ratePadding3
            val dy3: Double = Math.abs(yCenter - y3) / ratePadding3
            val dx4: Double = Math.abs(xCenter - x4) / ratePadding4
            val dy4: Double = Math.abs(yCenter - y4) / ratePadding4

            x1 -= dx1
            y1 -= dy1 * 1.8
            x2 += dx2
            y2 -= dy2 * 1.8
            x3 += dx3
            y3 += dy3 * 1.8
            x4 -= dx4
            y4 += dy4 * 1.8

            points[0].point.x = x1
            points[0].point.y = y1
            points[1].point.x = x2
            points[1].point.y = y2
            points[2].point.x = x3
            points[2].point.y = y3
            points[3].point.x = x4
            points[3].point.y = y4
        }
    }
}

class IDPoint(var vec: Double, var point: Point)