package com.Ekyc.ekyc2022.objectdetection

import android.content.Context
import android.graphics.Bitmap
import io.reactivex.Single
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import org.opencv.core.Point
import org.tensorflow.lite.support.image.TensorImage
import com.Ekyc.ekyc2022.global.Constants
import com.Ekyc.ekyc2022.global.ImageHelper
import com.Ekyc.ekyc2022.global.Utils.releaseBitmap
import com.Ekyc.ekyc2022.ml.IdCardModel
import java.security.InvalidParameterException
import kotlin.math.hypot

@Suppress("UNUSED")
class CardCornersDetectionProcessor(
    context: Context,
    private val callback: Callback,
    private val borderMarginHorizontal: Float = 0f,
    private val borderMarginVertical: Float = 0f,
    private val isFrontIdCard: Boolean = true,
    private val enableAutoCapture: Boolean = false
) {
    private val model = IdCardModel.newInstance(context)
    private val compositeDisposable = CompositeDisposable()

    @Volatile
    private var hasClosedModel = false

    @Volatile
    private var isProcessing = false

    companion object {
        private const val SCORE_THRESHOLD = 0.5f
        private const val CATEGORY_CORNER = "corner"
        private const val CATEGORY_EMBLEM = "symbol"
        private const val CATEGORY_FINGERPRINT = "finger"
    }

    fun process(bitmap: Bitmap?, takeAPhoto: Boolean) {
        if (isProcessing || bitmap == null || hasClosedModel) {
            callback.onAnalyzedBitmap(null, takeAPhoto)
            return
        }
        isProcessing = true
        if (enableAutoCapture) {
            compositeDisposable.add(
                alignIdCardIfMeetRequirements(bitmap)
                    .subscribeOn(Schedulers.computation())
                    .observeOn(AndroidSchedulers.mainThread())
                    .doOnSubscribe { isProcessing = true }
                    .subscribe({
                        if (it != null) {
                            callback.onDetectedIdCardBitmap(it)
                        }
                        isProcessing = false
                    }, {
                        bitmap.releaseBitmap()
                        isProcessing = false
                    })
            )
        } else {
            compositeDisposable.add(
                analyzeBitmap(bitmap)
                    .subscribeOn(Schedulers.computation())
                    .observeOn(AndroidSchedulers.mainThread())
                    .doOnSubscribe { isProcessing = true }
                    .subscribe({

                        val idCardInfo = getIdCardInfoFromBitmap(bitmap)

                        val alignedBitmap = alignBitmapIfMeetRequirements(it, takeAPhoto)
                        if (alignedBitmap == null) {
                            callback.onAnalyzedBitmap(null, takeAPhoto)
                        } else {
                            callback.onAnalyzedBitmap(
                                AlignCardResult(
                                    alignedBitmap,
                                    idCardInfo.type
                                ), takeAPhoto
                            )
                        }
                        isProcessing = false
                    }, {
                        bitmap.releaseBitmap()
                        isProcessing = false
                    })
            )
        }
    }

    @SuppressWarnings("WeakerAccess")
    fun alignIdCardIfMeetRequirements(bitmap: Bitmap): Single<AlignCardResult> {
        return Single.create { emitter ->
            try {
                if (hasClosedModel) {
                    emitter.onError(IllegalStateException("Model has been closed!"))
                } else {
                    if (ImageHelper.isBlurredImage(bitmap)) {
                        bitmap.releaseBitmap()
                        emitter.onError(InvalidParameterException("Bitmap is blurred"))
                    } else {
                        val idCardInfo = getIdCardInfoFromBitmap(bitmap)
                        val alignedBitmap = alignBitmapIfMeetRequirements(idCardInfo, false)
                        if (alignedBitmap == null) {
                            bitmap.releaseBitmap()
                            emitter.onError(InvalidParameterException("Id card not found"))
                        } else {
                            emitter.onSuccess(AlignCardResult(alignedBitmap, idCardInfo.type))
                        }
                    }
                }
            } catch (e: Exception) {
                bitmap.releaseBitmap()
                emitter.onError(e)
            }
        }
    }

    private fun analyzeBitmap(bitmap: Bitmap): Single<IdCardInfo> {
        return Single.create { emitter ->
            try {
                if (hasClosedModel) {
                    emitter.onError(IllegalStateException("Model has been closed!"))
                } else {
                    val idCardInfo = getIdCardInfoFromBitmap(bitmap)
                    emitter.onSuccess(idCardInfo)
                }
            } catch (e: Exception) {
                bitmap.releaseBitmap()
                emitter.onError(e)
            }
        }
    }

    private fun getIdCardInfoFromBitmap(bitmap: Bitmap): IdCardInfo {
        val image = TensorImage.fromBitmap(bitmap)
        val maxBorderX = bitmap.width - borderMarginHorizontal
        val maxBorderY = bitmap.height - borderMarginVertical
        val idCardInfo = IdCardInfo(bitmap)

        val outputs = model.process(image)
        for (detectedPoint in outputs.detectionResultList) {
            val location = detectedPoint.locationAsRectF
            val score = detectedPoint.scoreAsFloat
            val category = detectedPoint.categoryAsString

            if (!mapIdCardType(category, idCardInfo)) {
                if (score > SCORE_THRESHOLD) {
                    val point = Point(
                        ((location.left + location.right) / 2).toDouble(),
                        ((location.top + location.bottom) / 2).toDouble()
                    )

                    if (point.x > borderMarginHorizontal && point.x < maxBorderX
                        && point.y > borderMarginVertical && point.y < maxBorderY
                    ) {
                        when (category) {
                            CATEGORY_CORNER -> idCardInfo.cardCorners.add(point)
                            CATEGORY_EMBLEM -> idCardInfo.emblem = point
                            CATEGORY_FINGERPRINT -> idCardInfo.fingerprintPoints.add(
                                point
                            )
                        }
                    }
                }
            }
        }
        return idCardInfo
    }

    private fun mapIdCardType(category: String, idCardInfo: IdCardInfo): Boolean {
        when (category) {
            IdCardType.FRONT_IDENTITY_CARD,
            IdCardType.BACK_IDENTITY_CARD,
            IdCardType.FRONT_CITIZEN_IDENTIFICATION,
            IdCardType.BACK_CITIZEN_IDENTIFICATION,
            IdCardType.FRONT_CHIP_BASED_CITIZEN_IDENTIFICATION,
            IdCardType.BACK_CHIP_BASED_CITIZEN_IDENTIFICATION -> {
                idCardInfo.type = category
                return true
            }
        }
        return false
    }

    private fun alignBitmapIfMeetRequirements(idCardInfo: IdCardInfo, takeAPhoto: Boolean): Bitmap? {
        if (idCardInfo.checkHasEnoughInfo() && checkValidDistance(idCardInfo)) {
            val listPoint = ArrayList<Point>()
            listPoint.addAll(idCardInfo.cardCorners)
            listPoint.addAll(idCardInfo.fingerprintPoints)
            listPoint.add(idCardInfo.emblem!!)
            return if ((isFrontIdCard && listPoint.size == AlignIdImageHelper.FRONT_ID_CARD_POINTS) ||
                (!isFrontIdCard && listPoint.size == AlignIdImageHelper.BACK_ID_CARD_POINTS)
            ) {
                val alignedBitmap =
                    AlignIdImageHelper.alignIDCard(idCardInfo.bitmap, listPoint, idCardInfo.type!!)
                alignedBitmap
            } else {
                null
            }
        } else {
            return null
        }
    }

    /**
     * Check if the distance between camera and id card is close enough.
     */
    private fun checkValidDistance(idCardInfo: IdCardInfo): Boolean {
        if (idCardInfo.cardCorners.size < Constants.CARD_CORNER_COUNT) {
            println("----226checkValidDistance fale")
            return false
        }
        val bitmapArea = idCardInfo.bitmap.width * idCardInfo.bitmap.height
        val firstPoint = idCardInfo.cardCorners[0]
        val distance1 = hypot(
            firstPoint.x - idCardInfo.cardCorners[1].x,
            firstPoint.y - idCardInfo.cardCorners[1].y,
        )
        val distance2 = hypot(
            firstPoint.x - idCardInfo.cardCorners[2].x,
            firstPoint.y - idCardInfo.cardCorners[2].y,
        )
        val distance3 = hypot(
            firstPoint.x - idCardInfo.cardCorners[3].x,
            firstPoint.y - idCardInfo.cardCorners[3].y,
        )
        val cardArea = minOf(distance1 * distance2, distance2 * distance3, distance3 * distance1)
        return (cardArea / bitmapArea > 0.3)
    }

    fun close() {
        model.close()
        hasClosedModel = true
        compositeDisposable.dispose()
    }

    interface Callback {
        fun onDetectedIdCardBitmap(result: AlignCardResult)

        fun onAnalyzedBitmap(result: AlignCardResult?, takeAPhoto: Boolean)
    }
}